@@ -4,13 +4,11 @@
 # See LICENSE.txt for permissions.
 #++
 
-require 'fileutils'
-require 'pathname'
-require 'rbconfig'
-
 require 'rubygems/format'
+require 'rubygems/exceptions'
 require 'rubygems/ext'
 require 'rubygems/require_paths_builder'
+require 'rubygems/user_interaction'
 
 ##
 # The installer class processes RubyGem .gem files and installs the files
@@ -21,7 +19,7 @@ require 'rubygems/require_paths_builder'
 # gemspec in the specifications dir, storing the cached gem in the cache dir,
 # and installing either wrappers or symlinks for executables.
 #
-# The installer fires pre and post install hooks.  Hooks can be added either
+# The installer invokes pre and post install hooks.  Hooks can be added either
 # through a rubygems_plugin.rb file in an installed gem or via a
 # rubygems/defaults/#{RUBY_ENGINE}.rb or rubygems/defaults/operating_system.rb
 # file.  See Gem.pre_install and Gem.post_install for details.
@@ -41,7 +39,9 @@ class Gem::Installer
 
   include Gem::UserInteraction
 
-  include Gem::RequirePathsBuilder
+  include Gem::RequirePathsBuilder if Gem::QUICKLOADER_SUCKAGE
+
+  attr_reader :gem
 
   ##
   # The directory a gem's executables will be installed into
@@ -54,20 +54,14 @@ class Gem::Installer
   attr_reader :gem_home
 
   ##
-  # The Gem::Specification for the gem being installed
+  # The options passed when the Gem::Installer was instantiated.
 
-  attr_reader :spec
+  attr_reader :options
 
-  @home_install_warning = false
   @path_warning = false
 
   class << self
 
-    ##
-    # True if we've warned about ~/.gems install
-
-    attr_accessor :home_install_warning
-
     ##
     # True if we've warned about PATH not including Gem.bindir
 
@@ -98,75 +92,41 @@ class Gem::Installer
   # :wrappers:: Install wrappers if true, symlinks if false.
 
   def initialize(gem, options={})
+    require 'fileutils'
+
     @gem = gem
-
-    options = {
-      :bin_dir      => nil,
-      :env_shebang  => false,
-      :exec_format  => false,
-      :force        => false,
-      :install_dir  => Gem.dir,
-      :source_index => Gem.source_index,
-    }.merge options
-
-    @env_shebang         = options[:env_shebang]
-    @force               = options[:force]
-    gem_home             = options[:install_dir]
-    @gem_home            = Pathname.new(gem_home).expand_path
-    @ignore_dependencies = options[:ignore_dependencies]
-    @format_executable   = options[:format_executable]
-    @security_policy     = options[:security_policy]
-    @wrappers            = options[:wrappers]
-    @bin_dir             = options[:bin_dir]
-    @development         = options[:development]
-    @source_index        = options[:source_index]
-
-    begin
-      @format = Gem::Format.from_file_by_path @gem, @security_policy
-    rescue Gem::Package::FormatError
-      raise Gem::InstallError, "invalid gem format for #{@gem}"
-    end
-
-    begin
-      FileUtils.mkdir_p @gem_home
-    rescue Errno::EACCES, Errno::ENOTDIR
-      # We'll divert to ~/.gems below
-    end
-
-    if not File.writable? @gem_home or
-        # TODO: Shouldn't have to test for existence of bindir; tests need it.
-        (@gem_home.to_s == Gem.dir and File.exist? Gem.bindir and
-         not File.writable? Gem.bindir) then
-      if options[:user_install] == false then # You don't want to use ~
-        raise Gem::FilePermissionError, @gem_home
-      elsif options[:user_install].nil? then
-        unless self.class.home_install_warning or options[:unpack] then
-          alert_warning "Installing to ~/.gem since #{@gem_home} and\n\t  #{Gem.bindir} aren't both writable."
-          self.class.home_install_warning = true
-        end
-      end
-      options[:user_install] = true
-    end
+    @options = options
+    process_options
 
     if options[:user_install] and not options[:unpack] then
       @gem_home = Gem.user_dir
-
-      user_bin_dir = File.join(@gem_home, 'bin')
-      unless ENV['PATH'].split(File::PATH_SEPARATOR).include? user_bin_dir then
-        unless self.class.path_warning then
-          alert_warning "You don't have #{user_bin_dir} in your PATH,\n\t  gem executables will not run."
-          self.class.path_warning = true
-        end
-      end
-
-      FileUtils.mkdir_p @gem_home unless File.directory? @gem_home
-      # If it's still not writable, you've got issues.
-      raise Gem::FilePermissionError, @gem_home unless File.writable? @gem_home
+      check_that_user_bin_dir_is_in_path
     end
+  end
 
-    @spec = @format.spec
+  ##
+  # Lazy accessor for the spec's gem directory.
 
-    @gem_dir = File.join(@gem_home, "gems", @spec.full_name).untaint
+  def gem_dir
+    @gem_dir ||= spec.gem_dir.dup.untaint
+  end
+
+  ##
+  # Lazy accessor for the installer's Gem::Format instance.
+
+  def format
+    begin
+      @format ||= Gem::Format.from_file_by_path gem, @security_policy
+    rescue Gem::Package::FormatError
+      raise Gem::InstallError, "invalid gem format for #{gem}"
+    end
+  end
+
+  ##
+  # Lazy accessor for the installer's spec.
+
+  def spec
+    @spec ||= format.spec
   end
 
   ##
@@ -181,72 +141,83 @@ class Gem::Installer
   #     specifications/<gem-version>.gemspec #=> the Gem::Specification
 
   def install
+    current_home = Gem.dir
+    current_path = Gem.paths.path
+
+    verify_gem_home(options[:unpack])
+    Gem.use_paths gem_home, current_path # HACK: shouldn't need Gem.paths.path
+
     # If we're forcing the install then disable security unless the security
-    # policy says that we only install singed gems.
+    # policy says that we only install signed gems.
     @security_policy = nil if @force and @security_policy and
                               not @security_policy.only_signed
 
-    unless @force then
-      if rrv = @spec.required_ruby_version then
-        unless rrv.satisfied_by? Gem.ruby_version then
-          raise Gem::InstallError, "#{@spec.name} requires Ruby version #{rrv}"
-        end
-      end
-
-      if rrgv = @spec.required_rubygems_version then
-        unless rrgv.satisfied_by? Gem::Version.new(Gem::RubyGemsVersion) then
-          raise Gem::InstallError,
-                "#{@spec.name} requires RubyGems version #{rrgv}"
-        end
-      end
-
-      unless @ignore_dependencies then
-        deps = @spec.runtime_dependencies
-        deps |= @spec.development_dependencies if @development
-
-        deps.each do |dep_gem|
-          ensure_dependency @spec, dep_gem
-        end
-      end
+    unless @force
+      ensure_required_ruby_version_met
+      ensure_required_rubygems_version_met
+      ensure_dependencies_met unless @ignore_dependencies
     end
 
     Gem.pre_install_hooks.each do |hook|
-      hook.call self
+      result = hook.call self
+
+      if result == false then
+        location = " at #{$1}" if hook.inspect =~ /@(.*:\d+)/
+
+        message = "pre-install hook#{location} failed for #{spec.full_name}"
+        raise Gem::InstallError, message
+      end
     end
 
-    FileUtils.mkdir_p @gem_home unless File.directory? @gem_home
+    Gem.ensure_gem_subdirectories gem_home
 
-    Gem.ensure_gem_subdirectories @gem_home
+    # Completely remove any previous gem files
+    FileUtils.rm_rf(gem_dir) if File.exist? gem_dir
 
-    FileUtils.mkdir_p @gem_dir
+    FileUtils.mkdir_p gem_dir
 
     extract_files
-    generate_bin
     build_extensions
-    write_spec
 
-    write_require_paths_file_if_needed
+    Gem.post_build_hooks.each do |hook|
+      result = hook.call self
 
-    # HACK remove?  Isn't this done in multiple places?
-    cached_gem = File.join @gem_home, "cache", @gem.split(/\//).pop
-    unless File.exist? cached_gem then
-      FileUtils.cp @gem, File.join(@gem_home, "cache")
+      if result == false then
+        FileUtils.rm_rf gem_dir
+
+        location = " at #{$1}" if hook.inspect =~ /@(.*:\d+)/
+
+        message = "post-build hook#{location} failed for #{spec.full_name}"
+        raise Gem::InstallError, message
+      end
     end
 
-    say @spec.post_install_message unless @spec.post_install_message.nil?
+    generate_bin
+    write_spec
 
-    @spec.loaded_from = File.join(@gem_home, 'specifications',
-                                  "#{@spec.full_name}.gemspec")
+    write_require_paths_file_if_needed if Gem::QUICKLOADER_SUCKAGE
 
-    @source_index.add_spec @spec
+    cache_file = spec.cache_file
+    FileUtils.cp gem, cache_file unless File.exist? cache_file
+
+    say spec.post_install_message unless spec.post_install_message.nil?
+
+    spec.loaded_from = spec.spec_file
+
+    Gem::Specification.add_spec spec unless Gem::Specification.include? spec
 
     Gem.post_install_hooks.each do |hook|
       hook.call self
     end
 
-    return @spec
+    return spec
   rescue Zlib::GzipFile::Error
-    raise Gem::InstallError, "gzip error installing #{@gem}"
+    raise Gem::InstallError, "gzip error installing #{gem}"
+  ensure
+    # conditional since we might be here because we're erroring out early.
+    if current_path
+      Gem.use_paths current_home, current_path
+    end
   end
 
   ##
@@ -260,7 +231,6 @@ class Gem::Installer
     unless installation_satisfies_dependency? dependency then
       raise Gem::InstallError, "#{spec.name} requires #{dependency}"
     end
-
     true
   end
 
@@ -268,7 +238,7 @@ class Gem::Installer
   # True if the gems in the source_index satisfy +dependency+.
 
   def installation_satisfies_dependency?(dependency)
-    @source_index.find_name(dependency.name, dependency.version_requirements).size > 0
+    not dependency.matching_specs.empty?
   end
 
   ##
@@ -276,7 +246,7 @@ class Gem::Installer
 
   def unpack(directory)
     @gem_dir = directory
-    @format = Gem::Format.from_file_by_path @gem, @security_policy
+    @format = Gem::Format.from_file_by_path gem, @security_policy
     extract_files
   end
 
@@ -285,22 +255,17 @@ class Gem::Installer
   # specifications directory.
 
   def write_spec
-    rubycode = @spec.to_ruby
-
-    file_name = File.join @gem_home, 'specifications',
-                          "#{@spec.full_name}.gemspec"
-
-    file_name.untaint
+    file_name = spec.spec_file.untaint
 
     File.open(file_name, "w") do |file|
-      file.puts rubycode
+      file.puts spec.to_ruby_for_cache
     end
   end
 
   ##
   # Creates windows .bat files for easy running of commands
 
-  def generate_windows_script(bindir, filename)
+  def generate_windows_script(filename, bindir)
     if Gem.win_platform? then
       script_name = filename + ".bat"
       script_path = File.join bindir, File.basename(script_name)
@@ -313,21 +278,27 @@ class Gem::Installer
   end
 
   def generate_bin
-    return if @spec.executables.nil? or @spec.executables.empty?
+    return if spec.executables.nil? or spec.executables.empty?
 
     # If the user has asked for the gem to be installed in a directory that is
     # the system gem directory, then use the system bin directory, else create
     # (or use) a new bin dir under the gem_home.
-    bindir = @bin_dir ? @bin_dir : Gem.bindir(@gem_home)
+    bindir = @bin_dir || Gem.bindir(gem_home)
 
     Dir.mkdir bindir unless File.exist? bindir
     raise Gem::FilePermissionError.new(bindir) unless File.writable? bindir
 
-    @spec.executables.each do |filename|
+    spec.executables.each do |filename|
       filename.untaint
-      bin_path = File.expand_path File.join(@gem_dir, @spec.bindir, filename)
+      bin_path = File.expand_path File.join(gem_dir, spec.bindir, filename)
+
+      unless File.exist? bin_path
+        warn "Hey?!?! Where did #{bin_path} go??"
+        next
+      end
+
       mode = File.stat(bin_path).mode | 0111
-      File.chmod mode, bin_path
+      FileUtils.chmod mode, bin_path
 
       if @wrappers then
         generate_bin_script filename, bindir
@@ -347,24 +318,15 @@ class Gem::Installer
   def generate_bin_script(filename, bindir)
     bin_script_path = File.join bindir, formatted_program_filename(filename)
 
-    exec_path = File.join @gem_dir, @spec.bindir, filename
+    FileUtils.rm_f bin_script_path # prior install may have been --no-wrappers
 
-    # HACK some gems don't have #! in their executables, restore 2008/06
-    #if File.read(exec_path, 2) == '#!' then
-      FileUtils.rm_f bin_script_path # prior install may have been --no-wrappers
+    File.open bin_script_path, 'wb', 0755 do |file|
+      file.print app_script_text(filename)
+    end
 
-      File.open bin_script_path, 'w', 0755 do |file|
-        file.print app_script_text(filename)
-      end
+    say bin_script_path if Gem.configuration.really_verbose
 
-      say bin_script_path if Gem.configuration.really_verbose
-
-      generate_windows_script bindir, filename
-    #else
-    #  FileUtils.rm_f bin_script_path
-    #  FileUtils.cp exec_path, bin_script_path,
-    #               :verbose => Gem.configuration.really_verbose
-    #end
+    generate_windows_script filename, bindir
   end
 
   ##
@@ -378,14 +340,14 @@ class Gem::Installer
       return
     end
 
-    src = File.join @gem_dir, 'bin', filename
+    src = File.join gem_dir, spec.bindir, filename
     dst = File.join bindir, formatted_program_filename(filename)
 
     if File.exist? dst then
       if File.symlink? dst then
         link = File.readlink(dst).split File::SEPARATOR
         cur_version = Gem::Version.create(link[-3].sub(/^.*-/, ''))
-        return if @spec.version < cur_version
+        return if spec.version < cur_version
       end
       File.unlink dst
     end
@@ -399,7 +361,7 @@ class Gem::Installer
 
   def shebang(bin_file_name)
     ruby_name = Gem::ConfigMap[:ruby_install_name] if @env_shebang
-    path = File.join @gem_dir, @spec.bindir, bin_file_name
+    path = spec.bin_file bin_file_name
     first_line = File.open(path, "rb") {|file| file.gets}
 
     if /\A#!/ =~ first_line then
@@ -420,16 +382,82 @@ class Gem::Installer
     end
   end
 
+  def ensure_required_ruby_version_met
+    if rrv = spec.required_ruby_version then
+      unless rrv.satisfied_by? Gem.ruby_version then
+        raise Gem::InstallError, "#{spec.name} requires Ruby version #{rrv}."
+      end
+    end
+  end
+
+  def ensure_required_rubygems_version_met
+    if rrgv = spec.required_rubygems_version then
+      unless rrgv.satisfied_by? Gem::Version.new(Gem::VERSION) then
+        raise Gem::InstallError,
+          "#{spec.name} requires RubyGems version #{rrgv}. " +
+          "Try 'gem update --system' to update RubyGems itself."
+      end
+    end
+  end
+
+  def ensure_dependencies_met
+    deps = spec.runtime_dependencies
+    deps |= spec.development_dependencies if @development
+
+    deps.each do |dep_gem|
+      ensure_dependency spec, dep_gem
+    end
+  end
+
+  def process_options
+    @options = {
+      :bin_dir      => nil,
+      :env_shebang  => false,
+      :exec_format  => false,
+      :force        => false,
+      :install_dir  => Gem.dir,
+    }.merge options
+
+    @env_shebang         = options[:env_shebang]
+    @force               = options[:force]
+    @gem_home            = options[:install_dir]
+    @ignore_dependencies = options[:ignore_dependencies]
+    @format_executable   = options[:format_executable]
+    @security_policy     = options[:security_policy]
+    @wrappers            = options[:wrappers]
+    @bin_dir             = options[:bin_dir]
+    @development         = options[:development]
+
+    raise "NOTE: Installer option :source_index is dead" if
+      options[:source_index]
+  end
+
+  def check_that_user_bin_dir_is_in_path
+    user_bin_dir = @bin_dir || Gem.bindir(gem_home)
+    unless ENV['PATH'].split(File::PATH_SEPARATOR).include? user_bin_dir then
+      unless self.class.path_warning then
+        alert_warning "You don't have #{user_bin_dir} in your PATH,\n\t  gem executables will not run."
+        self.class.path_warning = true
+      end
+    end
+  end
+
+  def verify_gem_home(unpack = false)
+    FileUtils.mkdir_p gem_home
+    raise Gem::FilePermissionError, gem_home unless
+      unpack or File.writable?(gem_home)
+  end
+
   ##
   # Return the text for an application file.
 
   def app_script_text(bin_file_name)
-    <<-TEXT
+    return <<-TEXT
 #{shebang bin_file_name}
 #
 # This file was generated by RubyGems.
 #
-# The application '#{@spec.name}' is installed as part of a gem, and
+# The application '#{spec.name}' is installed as part of a gem, and
 # this file is here to facilitate running it.
 #
 
@@ -442,8 +470,8 @@ if ARGV.first =~ /^_(.*)_$/ and Gem::Version.correct? $1 then
   ARGV.shift
 end
 
-gem '#{@spec.name}', version
-load Gem.bin_path('#{@spec.name}', '#{bin_file_name}', version)
+gem '#{spec.name}', version
+load Gem.bin_path('#{spec.name}', '#{bin_file_name}', version)
 TEXT
   end
 
@@ -451,14 +479,16 @@ TEXT
   # return the stub script text used to launch the true ruby script
 
   def windows_stub_script(bindir, bin_file_name)
-    <<-TEXT
+    ruby = File.basename(Gem.ruby).chomp('"')
+    return <<-TEXT
 @ECHO OFF
 IF NOT "%~f0" == "~f0" GOTO :WinNT
-@"#{File.basename(Gem.ruby).chomp('"')}" "#{File.join(bindir, bin_file_name)}" %1 %2 %3 %4 %5 %6 %7 %8 %9
+@"#{ruby}" "#{File.join(bindir, bin_file_name)}" %1 %2 %3 %4 %5 %6 %7 %8 %9
 GOTO :EOF
 :WinNT
-@"#{File.basename(Gem.ruby).chomp('"')}" "%~dpn0" %*
+@"#{ruby}" "%~dpn0" %*
 TEXT
+
   end
 
   ##
@@ -466,13 +496,12 @@ TEXT
   # configure scripts and rakefiles or mkrf_conf files.
 
   def build_extensions
-    return if @spec.extensions.empty?
+    return if spec.extensions.empty?
     say "Building native extensions.  This could take a while..."
-    start_dir = Dir.pwd
-    dest_path = File.join @gem_dir, @spec.require_paths.first
+    dest_path = File.join gem_dir, spec.require_paths.first
     ran_rake = false # only run rake once
 
-    @spec.extensions.each do |extension|
+    spec.extensions.each do |extension|
       break if ran_rake
       results = []
 
@@ -489,29 +518,37 @@ TEXT
                   nil
                 end
 
+
+      extension_dir = begin
+                        File.join gem_dir, File.dirname(extension)
+                      rescue TypeError # extension == nil
+                        gem_dir
+                      end
+
+
       begin
-        Dir.chdir File.join(@gem_dir, File.dirname(extension))
-        results = builder.build(extension, @gem_dir, dest_path, results)
+        Dir.chdir extension_dir do
+          results = builder.build(extension, gem_dir, dest_path, results)
 
-        say results.join("\n") if Gem.configuration.really_verbose
-
-      rescue => ex
+          say results.join("\n") if Gem.configuration.really_verbose
+        end
+      rescue
         results = results.join "\n"
 
-        File.open('gem_make.out', 'wb') { |f| f.puts results }
+        gem_make_out = File.join extension_dir, 'gem_make.out'
+
+        open gem_make_out, 'wb' do |io| io.puts results end
 
         message = <<-EOF
 ERROR: Failed to build gem native extension.
 
-#{results}
+        #{results}
 
-Gem files will remain installed in #{@gem_dir} for inspection.
-Results logged to #{File.join(Dir.pwd, 'gem_make.out')}
-        EOF
+Gem files will remain installed in #{gem_dir} for inspection.
+Results logged to #{gem_make_out}
+EOF
 
         raise ExtensionBuildError, message
-      ensure
-        Dir.chdir start_dir
       end
     end
   end
@@ -522,28 +559,27 @@ Results logged to #{File.join(Dir.pwd, 'gem_make.out')}
   # Ensures that files can't be installed outside the gem directory.
 
   def extract_files
-    expand_and_validate_gem_dir
-
     raise ArgumentError, "format required to extract from" if @format.nil?
 
     @format.file_entries.each do |entry, file_data|
       path = entry['path'].untaint
 
-      if path =~ /\A\// then # for extra sanity
-        raise Gem::InstallError,
-              "attempt to install file into #{entry['path'].inspect}"
+      if path.start_with? "/" then # for extra sanity
+        raise Gem::InstallError, "attempt to install file into #{entry['path']}"
       end
 
-      path = File.expand_path File.join(@gem_dir, path)
+      path = File.expand_path File.join(gem_dir, path)
 
-      if path !~ /\A#{Regexp.escape @gem_dir}/ then
-        msg = "attempt to install file into %p under %p" %
-                [entry['path'], @gem_dir]
+      unless path.start_with? gem_dir then
+        msg = "attempt to install file into %p under %s" %
+                [entry['path'], gem_dir]
         raise Gem::InstallError, msg
       end
 
-      FileUtils.rm_rf(path) if File.exists?(path)
-      FileUtils.mkdir_p File.dirname(path)
+      FileUtils.rm_rf(path) if File.exist? path
+
+      dir = File.dirname path
+      FileUtils.mkdir_p dir unless File.exist? dir
 
       File.open(path, "wb") do |out|
         out.write file_data
@@ -566,24 +602,14 @@ Results logged to #{File.join(Dir.pwd, 'gem_make.out')}
     end
   end
 
-  private
-
   ##
-  # HACK Pathname is broken on windows.
+  #
+  # Return the target directory where the gem is to be installed. This
+  # directory is not guaranteed to be populated.
+  #
 
-  def absolute_path? pathname
-    pathname.absolute? or (Gem.win_platform? and pathname.to_s =~ /\A[a-z]:/i)
+  def dir
+    gem_dir.to_s
   end
-
-  def expand_and_validate_gem_dir
-    @gem_dir = Pathname.new(@gem_dir).expand_path
-
-    unless absolute_path?(@gem_dir) then # HACK is this possible after #expand_path?
-      raise ArgumentError, "install directory %p not absolute" % @gem_dir
-    end
-
-    @gem_dir = @gem_dir.to_s
-  end
-
 end
 
