@@ -2,7 +2,7 @@
 
   math.c -
 
-  $Author: akr $
+  $Author: yugui $
   created at: Tue Jan 25 14:12:56 JST 1994
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -10,77 +10,73 @@
 **********************************************************************/
 
 #include "ruby/ruby.h"
+#include "internal.h"
 #include <math.h>
 #include <errno.h>
 
+#if defined(HAVE_SIGNBIT) && defined(__GNUC__) && defined(__sun__)
+    extern int signbit(double x);
+#endif
+
+#define numberof(array) (int)(sizeof(array) / sizeof((array)[0]))
+
 VALUE rb_mMath;
+VALUE rb_eMathDomainError;
 
-extern VALUE rb_to_float(VALUE val);
 #define Need_Float(x) do {if (TYPE(x) != T_FLOAT) {(x) = rb_to_float(x);}} while(0)
 #define Need_Float2(x,y) do {\
     Need_Float(x);\
     Need_Float(y);\
 } while (0)
 
-static void
-domain_check(double x, double y, const char *msg)
-{
-    while(1) {
-	if (errno) {
-	    rb_sys_fail(msg);
-	}
-	if (isnan(y)) {
-	    if (isnan(x)) break;
-#if defined(EDOM)
-	    errno = EDOM;
-#elif defined(ERANGE)
-	    errno = ERANGE;
-#endif
-	    continue;
-	}
-	break;
-    }
-}
-
-static void
-infinity_check(VALUE arg, double res, const char *msg)
-{
-    while(1) {
-	if (errno) {
-	    rb_sys_fail(msg);
-	}
-	if (isinf(res) && !isinf(RFLOAT_VALUE(arg))) {
-#if defined(EDOM)
-	    errno = EDOM;
-#elif defined(ERANGE)
-	    errno = ERANGE;
-#endif
-	    continue;
-	}
-	break;
-    }
-}
+#define domain_error(msg) \
+    rb_raise(rb_eMathDomainError, "Numerical argument is out of domain - " #msg);
 
 /*
  *  call-seq:
- *     Math.atan2(y, x)  => float
+ *     Math.atan2(y, x)  -> float
  *
  *  Computes the arc tangent given <i>y</i> and <i>x</i>. Returns
  *  -PI..PI.
  *
+ *    Math.atan2(-0.0, -1.0) #=> -3.141592653589793
+ *    Math.atan2(-1.0, -1.0) #=> -2.356194490192345
+ *    Math.atan2(-1.0, 0.0)  #=> -1.5707963267948966
+ *    Math.atan2(-1.0, 1.0)  #=> -0.7853981633974483
+ *    Math.atan2(-0.0, 1.0)  #=> -0.0
+ *    Math.atan2(0.0, 1.0)   #=> 0.0
+ *    Math.atan2(1.0, 1.0)   #=> 0.7853981633974483
+ *    Math.atan2(1.0, 0.0)   #=> 1.5707963267948966
+ *    Math.atan2(1.0, -1.0)  #=> 2.356194490192345
+ *    Math.atan2(0.0, -1.0)  #=> 3.141592653589793
+ *
  */
 
 static VALUE
 math_atan2(VALUE obj, VALUE y, VALUE x)
 {
+#ifndef M_PI
+# define M_PI 3.14159265358979323846
+#endif
+    double dx, dy;
     Need_Float2(y, x);
-    return DBL2NUM(atan2(RFLOAT_VALUE(y), RFLOAT_VALUE(x)));
+    dx = RFLOAT_VALUE(x);
+    dy = RFLOAT_VALUE(y);
+    if (dx == 0.0 && dy == 0.0) {
+	if (!signbit(dx))
+	    return DBL2NUM(dy);
+        if (!signbit(dy))
+	    return DBL2NUM(M_PI);
+	return DBL2NUM(-M_PI);
+    }
+    if (isinf(dx) && isinf(dy)) domain_error("atan2");
+    return DBL2NUM(atan2(dy, dx));
 }
 
 
 /*
  *  call-seq:
- *     Math.cos(x)    => float
+ *     Math.cos(x)    -> float
  *
  *  Computes the cosine of <i>x</i> (expressed in radians). Returns
  *  -1..1.
@@ -95,7 +91,7 @@ math_cos(VALUE obj, VALUE x)
 
 /*
  *  call-seq:
- *     Math.sin(x)    => float
+ *     Math.sin(x)    -> float
  *
  *  Computes the sine of <i>x</i> (expressed in radians). Returns
  *  -1..1.
@@ -112,7 +108,7 @@ math_sin(VALUE obj, VALUE x)
 
 /*
  *  call-seq:
- *     Math.tan(x)    => float
+ *     Math.tan(x)    -> float
  *
  *  Returns the tangent of <i>x</i> (expressed in radians).
  */
@@ -127,7 +123,7 @@ math_tan(VALUE obj, VALUE x)
 
 /*
  *  call-seq:
- *     Math.acos(x)    => float
+ *     Math.acos(x)    -> float
  *
  *  Computes the arc cosine of <i>x</i>. Returns 0..PI.
  */
@@ -138,16 +134,16 @@ math_acos(VALUE obj, VALUE x)
     double d0, d;
 
     Need_Float(x);
-    errno = 0;
     d0 = RFLOAT_VALUE(x);
+    /* check for domain error */
+    if (d0 < -1.0 || 1.0 < d0) domain_error("acos");
     d = acos(d0);
-    domain_check(d0, d, "acos");
     return DBL2NUM(d);
 }
 
 /*
  *  call-seq:
- *     Math.asin(x)    => float
+ *     Math.asin(x)    -> float
  *
  *  Computes the arc sine of <i>x</i>. Returns -{PI/2} .. {PI/2}.
  */
@@ -158,16 +154,16 @@ math_asin(VALUE obj, VALUE x)
     double d0, d;
 
     Need_Float(x);
-    errno = 0;
     d0 = RFLOAT_VALUE(x);
+    /* check for domain error */
+    if (d0 < -1.0 || 1.0 < d0) domain_error("asin");
     d = asin(d0);
-    domain_check(d0, d, "asin");
     return DBL2NUM(d);
 }
 
 /*
  *  call-seq:
- *     Math.atan(x)    => float
+ *     Math.atan(x)    -> float
  *
  *  Computes the arc tangent of <i>x</i>. Returns -{PI/2} .. {PI/2}.
  */
@@ -189,7 +185,7 @@ cosh(double x)
 
 /*
  *  call-seq:
- *     Math.cosh(x)    => float
+ *     Math.cosh(x)    -> float
  *
  *  Computes the hyperbolic cosine of <i>x</i> (expressed in radians).
  */
@@ -212,7 +208,7 @@ sinh(double x)
 
 /*
  *  call-seq:
- *     Math.sinh(x)    => float
+ *     Math.sinh(x)    -> float
  *
  *  Computes the hyperbolic sine of <i>x</i> (expressed in
  *  radians).
@@ -235,7 +231,7 @@ tanh(double x)
 
 /*
  *  call-seq:
- *     Math.tanh()    => float
+ *     Math.tanh()    -> float
  *
  *  Computes the hyperbolic tangent of <i>x</i> (expressed in
  *  radians).
@@ -250,7 +246,7 @@ math_tanh(VALUE obj, VALUE x)
 
 /*
  *  call-seq:
- *     Math.acosh(x)    => float
+ *     Math.acosh(x)    -> float
  *
  *  Computes the inverse hyperbolic cosine of <i>x</i>.
  */
@@ -261,16 +257,16 @@ math_acosh(VALUE obj, VALUE x)
     double d0, d;
 
     Need_Float(x);
-    errno = 0;
     d0 = RFLOAT_VALUE(x);
+    /* check for domain error */
+    if (d0 < 1.0) domain_error("acosh");
     d = acosh(d0);
-    domain_check(d0, d, "acosh");
     return DBL2NUM(d);
 }
 
 /*
  *  call-seq:
- *     Math.asinh(x)    => float
+ *     Math.asinh(x)    -> float
  *
  *  Computes the inverse hyperbolic sine of <i>x</i>.
  */
@@ -284,7 +280,7 @@ math_asinh(VALUE obj, VALUE x)
 
 /*
  *  call-seq:
- *     Math.atanh(x)    => float
+ *     Math.atanh(x)    -> float
  *
  *  Computes the inverse hyperbolic tangent of <i>x</i>.
  */
@@ -295,17 +291,19 @@ math_atanh(VALUE obj, VALUE x)
     double d0, d;
 
     Need_Float(x);
-    errno = 0;
     d0 = RFLOAT_VALUE(x);
+    /* check for domain error */
+    if (d0 <  -1.0 || +1.0 <  d0) domain_error("atanh");
+    /* check for pole error */
+    if (d0 == -1.0) return DBL2NUM(-INFINITY);
+    if (d0 == +1.0) return DBL2NUM(+INFINITY);
     d = atanh(d0);
-    domain_check(d0, d, "atanh");
-    infinity_check(x, d, "atanh");
     return DBL2NUM(d);
 }
 
 /*
  *  call-seq:
- *     Math.exp(x)    => float
+ *     Math.exp(x)    -> float
  *
  *  Returns e**x.
  *
@@ -333,8 +331,8 @@ math_exp(VALUE obj, VALUE x)
 
 /*
  *  call-seq:
- *     Math.log(numeric)    => float
- *     Math.log(num,base)   => float
+ *     Math.log(numeric)    -> float
+ *     Math.log(num,base)   -> float
  *
  *  Returns the natural logarithm of <i>numeric</i>.
  *  If additional second argument is given, it will be the base
@@ -355,15 +353,16 @@ math_log(int argc, VALUE *argv)
 
     rb_scan_args(argc, argv, "11", &x, &base);
     Need_Float(x);
-    errno = 0;
     d0 = RFLOAT_VALUE(x);
+    /* check for domain error */
+    if (d0 < 0.0) domain_error("log");
+    /* check for pole error */
+    if (d0 == 0.0) return DBL2NUM(-INFINITY);
     d = log(d0);
     if (argc == 2) {
 	Need_Float(base);
 	d /= log(RFLOAT_VALUE(base));
     }
-    domain_check(d0, d, "log");
-    infinity_check(x, d, "log");
     return DBL2NUM(d);
 }
 
@@ -381,7 +380,7 @@ extern double log2(double);
 
 /*
  *  call-seq:
- *     Math.log2(numeric)    => float
+ *     Math.log2(numeric)    -> float
  *
  *  Returns the base 2 logarithm of <i>numeric</i>.
  *
@@ -398,17 +397,18 @@ math_log2(VALUE obj, VALUE x)
     double d0, d;
 
     Need_Float(x);
-    errno = 0;
     d0 = RFLOAT_VALUE(x);
+    /* check for domain error */
+    if (d0 < 0.0) domain_error("log2");
+    /* check for pole error */
+    if (d0 == 0.0) return DBL2NUM(-INFINITY);
     d = log2(d0);
-    domain_check(d0, d, "log2");
-    infinity_check(x, d, "log2");
     return DBL2NUM(d);
 }
 
 /*
  *  call-seq:
- *     Math.log10(numeric)    => float
+ *     Math.log10(numeric)    -> float
  *
  *  Returns the base 10 logarithm of <i>numeric</i>.
  *
@@ -424,17 +424,18 @@ math_log10(VALUE obj, VALUE x)
     double d0, d;
 
     Need_Float(x);
-    errno = 0;
     d0 = RFLOAT_VALUE(x);
+    /* check for domain error */
+    if (d0 < 0.0) domain_error("log10");
+    /* check for pole error */
+    if (d0 == 0.0) return DBL2NUM(-INFINITY);
     d = log10(d0);
-    domain_check(d0, d, "log10");
-    infinity_check(x, d, "log10");
     return DBL2NUM(d);
 }
 
 /*
  *  call-seq:
- *     Math.sqrt(numeric)    => float
+ *     Math.sqrt(numeric)    -> float
  *
  *  Returns the non-negative square root of <i>numeric</i>.
  *
@@ -462,16 +463,17 @@ math_sqrt(VALUE obj, VALUE x)
     double d0, d;
 
     Need_Float(x);
-    errno = 0;
     d0 = RFLOAT_VALUE(x);
+    /* check for domain error */
+    if (d0 < 0.0) domain_error("sqrt");
+    if (d0 == 0.0) return DBL2NUM(0.0);
     d = sqrt(d0);
-    domain_check(d0, d, "sqrt");
     return DBL2NUM(d);
 }
 
 /*
  *  call-seq:
- *     Math.cbrt(numeric)    => float
+ *     Math.cbrt(numeric)    -> float
  *
  *  Returns the cube root of <i>numeric</i>.
  *
@@ -510,7 +512,7 @@ math_cbrt(VALUE obj, VALUE x)
 
 /*
  *  call-seq:
- *     Math.frexp(numeric)    => [ fraction, exponent ]
+ *     Math.frexp(numeric)    -> [ fraction, exponent ]
  *
  *  Returns a two-element array containing the normalized fraction (a
  *  <code>Float</code>) and exponent (a <code>Fixnum</code>) of
@@ -551,7 +553,7 @@ math_ldexp(VALUE obj, VALUE x, VALUE n)
 
 /*
  *  call-seq:
- *     Math.hypot(x, y)    => float
+ *     Math.hypot(x, y)    -> float
  *
  *  Returns sqrt(x**2 + y**2), the hypotenuse of a right-angled triangle
  *  with sides <i>x</i> and <i>y</i>.
@@ -568,7 +570,7 @@ math_hypot(VALUE obj, VALUE x, VALUE y)
 
 /*
  * call-seq:
- *    Math.erf(x)  => float
+ *    Math.erf(x)  -> float
  *
  *  Calculates the error function of x.
  */
@@ -582,7 +584,7 @@ math_erf(VALUE obj, VALUE x)
 
 /*
  * call-seq:
- *    Math.erfc(x)  => float
+ *    Math.erfc(x)  -> float
  *
  *  Calculates the complementary error function of x.
  */
@@ -596,12 +598,12 @@ math_erfc(VALUE obj, VALUE x)
 
 /*
  * call-seq:
- *    Math.gamma(x)  => float
+ *    Math.gamma(x)  -> float
  *
  *  Calculates the gamma function of x.
  *
  *  Note that gamma(n) is same as fact(n-1) for integer n > 0.
- *  However gamma(n) returns float and possibly has error in calculation.
+ *  However gamma(n) returns float and can be an approximation.
  *
  *   def fact(n) (1..n).inject(1) {|r,i| r*i } end
  *   1.upto(26) {|i| p [i, Math.gamma(i), fact(i-1)] }
@@ -637,7 +639,7 @@ math_erfc(VALUE obj, VALUE x)
 static VALUE
 math_gamma(VALUE obj, VALUE x)
 {
-    static double fact_table[] = {
+    static const double fact_table[] = {
         /* fact(0) */ 1.0,
         /* fact(1) */ 1.0,
         /* fact(2) */ 2.0,
@@ -662,27 +664,30 @@ math_gamma(VALUE obj, VALUE x)
         /* fact(21) */ 51090942171709440000.0,
         /* fact(22) */ 1124000727777607680000.0,
         /* fact(23)=25852016738884976640000 needs 56bit mantissa which is
-         * impossible to represent exactly in IEEE 754 double. */
+         * impossible to represent exactly in IEEE 754 double which have
+         * 53bit mantissa. */
     };
     double d0, d;
     double intpart, fracpart;
     Need_Float(x);
     d0 = RFLOAT_VALUE(x);
+    /* check for domain error */
+    if (isinf(d0) && signbit(d0)) domain_error("gamma");
     fracpart = modf(d0, &intpart);
-    if (fracpart == 0.0 &&
-            0 < intpart &&
-            (size_t)intpart <= sizeof(fact_table)/sizeof(*fact_table)) {
-        return DBL2NUM(fact_table[(int)intpart - 1]);
+    if (fracpart == 0.0) {
+	if (intpart < 0) domain_error("gamma");
+	if (0 < intpart &&
+	    intpart - 1 < (double)numberof(fact_table)) {
+	    return DBL2NUM(fact_table[(int)intpart - 1]);
+	}
     }
-    errno = 0;
     d = tgamma(d0);
-    domain_check(d0, d, "gamma");
     return DBL2NUM(d);
 }
 
 /*
  * call-seq:
- *    Math.lgamma(x)  => [float, -1 or 1]
+ *    Math.lgamma(x)  -> [float, -1 or 1]
  *
  *  Calculates the logarithmic gamma of x and
  *  the sign of gamma of x.
@@ -696,13 +701,16 @@ static VALUE
 math_lgamma(VALUE obj, VALUE x)
 {
     double d0, d;
-    int sign;
+    int sign=1;
     VALUE v;
     Need_Float(x);
-    errno = 0;
     d0 = RFLOAT_VALUE(x);
+    /* check for domain error */
+    if (isinf(d0)) {
+	if (signbit(d0)) domain_error("lgamma");
+	return rb_assoc_new(DBL2NUM(INFINITY), INT2FIX(1));
+    }
     d = lgamma_r(d0, &sign);
-    domain_check(d0, d, "lgamma");
     v = DBL2NUM(d);
     return rb_assoc_new(v, INT2FIX(sign));
 }
@@ -740,6 +748,24 @@ exp1(sqrt)
 
 
 /*
+ *  Document-class: Math::DomainError
+ *
+ *  Raised when a mathematical function is evaluated outside of its
+ *  domain of definition.
+ *
+ *  For example, since +cos+ returns values in the range -1..1,
+ *  its inverse function +acos+ is only defined on that interval:
+ *
+ *     Math.acos(42)
+ *
+ *  <em>produces:</em>
+ *
+ *     Math::DomainError: Numerical argument is out of domain - "acos"
+ */
+
+/*
+ *  Document-class: Math
+ *
  *  The <code>Math</code> module contains module functions for basic
  *  trigonometric and transcendental functions. See class
  *  <code>Float</code> for a list of constants that
@@ -751,6 +777,7 @@ void
 Init_Math(void)
 {
     rb_mMath = rb_define_module("Math");
+    rb_eMathDomainError = rb_define_class_under(rb_mMath, "DomainError", rb_eStandardError);
 
 #ifdef M_PI
     rb_define_const(rb_mMath, "PI", DBL2NUM(M_PI));
