@@ -2,7 +2,7 @@
 
   vm_insnhelper.c - instruction helper functions.
 
-  $Author: ko1 $
+  $Author: kosaki $
 
   Copyright (C) 2007 Koichi Sasada
 
@@ -11,6 +11,8 @@
 /* finish iseq array */
 #include "insns.inc"
 #include <math.h>
+#include "constant.h"
+#include "internal.h"
 
 /* control stack frame */
 
@@ -26,9 +28,13 @@ vm_push_frame(rb_thread_t * th, const rb_iseq_t * iseq,
 	      const VALUE *pc, VALUE *sp, VALUE *lfp,
 	      int local_size)
 {
-    rb_control_frame_t * const cfp = th->cfp = th->cfp - 1;
+    rb_control_frame_t * const cfp = th->cfp - 1;
     int i;
 
+    if ((void *)(sp + local_size) >= (void *)cfp) {
+	rb_exc_raise(sysstack_error);
+    }
+    th->cfp = cfp;
     /* setup vm value stack */
 
     /* nil initialize */
@@ -54,7 +60,9 @@ vm_push_frame(rb_thread_t * th, const rb_iseq_t * iseq,
     cfp->self = self;
     cfp->lfp = lfp;
     cfp->dfp = sp;
+    cfp->block_iseq = 0;
     cfp->proc = 0;
+    cfp->me = 0;
 
 #define COLLECT_PROFILE 0
 #if COLLECT_PROFILE
@@ -98,16 +106,42 @@ vm_pop_frame(rb_thread_t *th)
 
 /* method dispatch */
 
+NORETURN(static void argument_error(const rb_iseq_t *iseq, int miss_argc, int correct_argc));
+static void
+argument_error(const rb_iseq_t *iseq, int miss_argc, int correct_argc)
+{
+    VALUE mesg = rb_sprintf("wrong number of arguments (%d for %d)", miss_argc, correct_argc);
+    VALUE exc = rb_exc_new3(rb_eArgError, mesg);
+    VALUE bt = rb_make_backtrace();
+    VALUE err_line = 0;
+
+    if (iseq) {
+	int line_no = 1;
+
+	if (iseq->insn_info_size) {
+	    line_no = iseq->insn_info_table[0].line_no;
+	}
+
+	err_line = rb_sprintf("%s:%d:in `%s'",
+			      RSTRING_PTR(iseq->filename),
+			      line_no, RSTRING_PTR(iseq->name));
+	rb_funcall(bt, rb_intern("unshift"), 1, err_line);
+    }
+
+    rb_funcall(exc, rb_intern("set_backtrace"), 1, bt);
+    rb_exc_raise(exc);
+}
+
 #define VM_CALLEE_SETUP_ARG(ret, th, iseq, orig_argc, orig_argv, block) \
-    if (LIKELY(iseq->arg_simple & 0x01)) { \
+    if (LIKELY((iseq)->arg_simple & 0x01)) { \
 	/* simple check */ \
-	if (orig_argc != iseq->argc) { \
-	    rb_raise(rb_eArgError, "wrong number of arguments (%d for %d)", orig_argc, iseq->argc); \
+	if ((orig_argc) != (iseq)->argc) { \
+	    argument_error((iseq), (orig_argc), (iseq)->argc); \
 	} \
-	ret = 0; \
+	(ret) = 0; \
     } \
     else { \
-	ret = vm_callee_setup_arg_complex(th, iseq, orig_argc, orig_argv, block); \
+	(ret) = vm_callee_setup_arg_complex((th), (iseq), (orig_argc), (orig_argv), (block)); \
     }
 
 static inline int
@@ -124,8 +158,7 @@ vm_callee_setup_arg_complex(rb_thread_t *th, const rb_iseq_t * iseq,
 
     /* mandatory */
     if (argc < (m + iseq->arg_post_len)) { /* check with post arg */
-	rb_raise(rb_eArgError, "wrong number of arguments (%d for %d)",
-		 argc, m + iseq->arg_post_len);
+	argument_error(iseq, argc, m + iseq->arg_post_len);
     }
 
     argv += m;
@@ -148,8 +181,7 @@ vm_callee_setup_arg_complex(rb_thread_t *th, const rb_iseq_t * iseq,
 	const int opts = iseq->arg_opts - 1 /* no opt */;
 
 	if (iseq->arg_rest == -1 && argc > opts) {
-	    rb_raise(rb_eArgError, "wrong number of arguments (%d for %d)",
-		     orig_argc, m + opts + iseq->arg_post_len);
+	    argument_error(iseq, orig_argc, m + opts + iseq->arg_post_len);
 	}
 
 	if (argc > opts) {
@@ -179,8 +211,7 @@ vm_callee_setup_arg_complex(rb_thread_t *th, const rb_iseq_t * iseq,
 	const rb_block_t *blockptr = *block;
 
 	if (argc != 0) {
-	    rb_raise(rb_eArgError, "wrong number of arguments (%d for %d)",
-		     orig_argc, m + iseq->arg_post_len);
+	    argument_error(iseq, orig_argc, m + iseq->arg_post_len);
 	}
 
 	if (blockptr) {
@@ -355,44 +386,30 @@ call_cfunc(VALUE (*func)(), VALUE recv,
 }
 
 static inline VALUE
-vm_call_cfunc(rb_thread_t *th, rb_control_frame_t *reg_cfp,
-	      int num, VALUE recv, const rb_block_t *blockptr, VALUE flag,
+vm_call_cfunc(rb_thread_t *th, volatile rb_control_frame_t *reg_cfp,
+	      int num, volatile VALUE recv, const rb_block_t *blockptr,
 	      const rb_method_entry_t *me)
 {
-    VALUE val = 0;
-    int state = 0;
-    VALUE klass = me->klass;
-    ID id = me->original_id;
+    volatile VALUE val = 0;
+    const rb_method_definition_t *def = me->def;
+    rb_control_frame_t *cfp;
 
-    EXEC_EVENT_HOOK(th, RUBY_EVENT_C_CALL, recv, id, klass);
+    EXEC_EVENT_HOOK(th, RUBY_EVENT_C_CALL, recv, me->called_id, me->klass);
 
-    TH_PUSH_TAG(th);
-    // TODO: fix me.  separate event
-    if (th->event_flags & RUBY_EVENT_C_RETURN) {
-	state = TH_EXEC_TAG();
+    cfp = vm_push_frame(th, 0, VM_FRAME_MAGIC_CFUNC,
+			recv, (VALUE) blockptr, 0, reg_cfp->sp, 0, 1);
+    cfp->me = me;
+    reg_cfp->sp -= num + 1;
+
+    val = call_cfunc(def->body.cfunc.func, recv, (int)def->body.cfunc.argc, num, reg_cfp->sp + 1);
+
+    if (reg_cfp != th->cfp + 1) {
+	rb_bug("cfp consistency error - send");
     }
-    else {
-	_th->tag = _tag.prev;
-    }
-    if (state == 0) {
-	rb_control_frame_t *cfp =
-	    vm_push_frame(th, 0, VM_FRAME_MAGIC_CFUNC,
-			  recv, (VALUE) blockptr, 0, reg_cfp->sp, 0, 1);
 
-	cfp->me = me;
-	reg_cfp->sp -= num + 1;
+    vm_pop_frame(th);
 
-	val = call_cfunc(me->body.cfunc.func, recv, (int)me->body.cfunc.argc, num, reg_cfp->sp + 1);
-
-	if (reg_cfp != th->cfp + 1) {
-	    rb_bug("cfp consistency error - send");
-	}
-
-	vm_pop_frame(th);
-    }
-    TH_POP_TAG();
-    EXEC_EVENT_HOOK(th, RUBY_EVENT_C_RETURN, recv, id, klass);
-    if (state) TH_JUMP_TAG(th, state);
+    EXEC_EVENT_HOOK(th, RUBY_EVENT_C_RETURN, recv, me->called_id, me->klass);
 
     return val;
 }
@@ -401,15 +418,18 @@ static inline VALUE
 vm_call_bmethod(rb_thread_t *th, VALUE recv, int argc, const VALUE *argv,
 		const rb_block_t *blockptr, const rb_method_entry_t *me)
 {
-    rb_control_frame_t *cfp = th->cfp;
     rb_proc_t *proc;
     VALUE val;
 
-    /* control block frame */
-    (cfp-2)->me = me;
+    EXEC_EVENT_HOOK(th, RUBY_EVENT_CALL, recv, me->called_id, me->klass);
 
-    GetProcPtr(me->body.proc, proc);
+    /* control block frame */
+    th->passed_me = me;
+    GetProcPtr(me->def->body.proc, proc);
     val = rb_vm_invoke_proc(th, proc, recv, argc, argv, blockptr);
+
+    EXEC_EVENT_HOOK(th, RUBY_EVENT_RETURN, recv, me->called_id, me->klass);
+
     return val;
 }
 
@@ -435,13 +455,13 @@ vm_method_missing(rb_thread_t *th, ID id, VALUE recv,
 }
 
 static inline void
-vm_setup_method(rb_thread_t *th, rb_control_frame_t *cfp, 
-		VALUE recv, int argc, const rb_block_t *blockptr, VALUE flag, 
+vm_setup_method(rb_thread_t *th, rb_control_frame_t *cfp,
+		VALUE recv, int argc, const rb_block_t *blockptr, VALUE flag,
 		const rb_method_entry_t *me)
 {
     int opt_pc, i;
     VALUE *sp, *rsp = cfp->sp - argc;
-    rb_iseq_t *iseq = me->body.iseq;
+    rb_iseq_t *iseq = me->def->body.iseq;
 
     VM_CALLEE_SETUP_ARG(opt_pc, th, iseq, argc, rsp, &blockptr);
 
@@ -500,21 +520,21 @@ vm_call_method(rb_thread_t *th, rb_control_frame_t *cfp,
     if (me != 0) {
 	if ((me->flag == 0)) {
 	  normal_method_dispatch:
-	    switch (me->type) {
+	    switch (me->def->type) {
 	      case VM_METHOD_TYPE_ISEQ:{
 		vm_setup_method(th, cfp, recv, num, blockptr, flag, me);
 		return Qundef;
 	      }
 	      case VM_METHOD_TYPE_NOTIMPLEMENTED:
 	      case VM_METHOD_TYPE_CFUNC:{
-		val = vm_call_cfunc(th, cfp, num, recv, blockptr, flag, me);
+		val = vm_call_cfunc(th, cfp, num, recv, blockptr, me);
 		break;
 	      }
 	      case VM_METHOD_TYPE_ATTRSET:{
 		if (num != 1) {
 		    rb_raise(rb_eArgError, "wrong number of arguments (%d for 1)", num);
 		}
-		val = rb_ivar_set(recv, me->body.attr_id, *(cfp->sp - 1));
+		val = rb_ivar_set(recv, me->def->body.attr.id, *(cfp->sp - 1));
 		cfp->sp -= 2;
 		break;
 	      }
@@ -522,10 +542,18 @@ vm_call_method(rb_thread_t *th, rb_control_frame_t *cfp,
 		if (num != 0) {
 		    rb_raise(rb_eArgError, "wrong number of arguments (%d for 0)", num);
 		}
-		val = rb_attr_get(recv, me->body.attr_id);
+		val = rb_attr_get(recv, me->def->body.attr.id);
 		cfp->sp -= 1;
 		break;
 	      }
+	      case VM_METHOD_TYPE_MISSING:{
+		VALUE *argv = ALLOCA_N(VALUE, num+1);
+		argv[0] = ID2SYM(me->def->original_id);
+		MEMCPY(argv+1, cfp->sp - num, VALUE, num);
+		cfp->sp += - num - 1;
+		val = rb_funcall2(recv, rb_intern("method_missing"), num+1, argv);
+		break;
+	      }
 	      case VM_METHOD_TYPE_BMETHOD:{
 		VALUE *argv = ALLOCA_N(VALUE, num);
 		MEMCPY(argv, cfp->sp - num, VALUE, num);
@@ -545,49 +573,48 @@ vm_call_method(rb_thread_t *th, rb_control_frame_t *cfp,
 		}
 	      }
 	      case VM_METHOD_TYPE_OPTIMIZED:{
-		  switch (me->body.optimize_type) {
-		    case OPTIMIZED_METHOD_TYPE_SEND: {
-			rb_control_frame_t *reg_cfp = cfp;
-			rb_num_t i = num - 1;
-			VALUE sym;
+		switch (me->def->body.optimize_type) {
+		  case OPTIMIZED_METHOD_TYPE_SEND: {
+		    rb_control_frame_t *reg_cfp = cfp;
+		    rb_num_t i = num - 1;
+		    VALUE sym;
 
-			if (num == 0) {
-			    rb_raise(rb_eArgError, "no method name given");
-			}
-
-			sym = TOPN(i);
-			id = SYMBOL_P(sym) ? SYM2ID(sym) : rb_to_id(sym);
-			/* shift arguments */
-			if (i > 0) {
-			    MEMMOVE(&TOPN(i), &TOPN(i-1), VALUE, i);
-			}
-			me = rb_method_entry(CLASS_OF(recv), id);
-			num -= 1;
-			DEC_SP(1);
-			flag |= VM_CALL_FCALL_BIT;
-
-			goto start_method_dispatch;
+		    if (num == 0) {
+			rb_raise(rb_eArgError, "no method name given");
 		    }
-		      break;
-		    case OPTIMIZED_METHOD_TYPE_CALL: {
-			rb_proc_t *proc;
-			int argc = num;
-			VALUE *argv = ALLOCA_N(VALUE, num);
-			GetProcPtr(recv, proc);
-			MEMCPY(argv, cfp->sp - num, VALUE, num);
-			cfp->sp -= num + 1;
 
-			val = rb_vm_invoke_proc(th, proc, proc->block.self, argc, argv, blockptr);
+		    sym = TOPN(i);
+		    id = SYMBOL_P(sym) ? SYM2ID(sym) : rb_to_id(sym);
+		    /* shift arguments */
+		    if (i > 0) {
+			MEMMOVE(&TOPN(i), &TOPN(i-1), VALUE, i);
 		    }
-		      break;
-		    default:
-		      rb_bug("eval_invoke_method: unsupported optimized method type (%d)",
-			     me->body.optimize_type);
+		    me = rb_method_entry(CLASS_OF(recv), id);
+		    num -= 1;
+		    DEC_SP(1);
+		    flag |= VM_CALL_FCALL_BIT | VM_CALL_OPT_SEND_BIT;
+
+		    goto start_method_dispatch;
 		  }
-		  break;
+		  case OPTIMIZED_METHOD_TYPE_CALL: {
+		    rb_proc_t *proc;
+		    int argc = num;
+		    VALUE *argv = ALLOCA_N(VALUE, num);
+		    GetProcPtr(recv, proc);
+		    MEMCPY(argv, cfp->sp - num, VALUE, num);
+		    cfp->sp -= num + 1;
+
+		    val = rb_vm_invoke_proc(th, proc, proc->block.self, argc, argv, blockptr);
+		    break;
+		  }
+		  default:
+		    rb_bug("eval_invoke_method: unsupported optimized method type (%d)",
+			   me->def->body.optimize_type);
+		}
+		break;
 	      }
 	      default:{
-		rb_bug("eval_invoke_method: unsupported method type (%d)", me->type);
+		rb_bug("eval_invoke_method: unsupported method type (%d)", me->def->type);
 		break;
 	      }
 	    }
@@ -604,14 +631,14 @@ vm_call_method(rb_thread_t *th, rb_control_frame_t *cfp,
 		}
 		val = vm_method_missing(th, id, recv, num, blockptr, stat);
 	    }
-	    else if ((me->flag & NOEX_MASK) & NOEX_PROTECTED) {
+	    else if (!(flag & VM_CALL_OPT_SEND_BIT) && (me->flag & NOEX_MASK) & NOEX_PROTECTED) {
 		VALUE defined_class = me->klass;
 
-		if (TYPE(defined_class) == T_ICLASS) {
+		if (RB_TYPE_P(defined_class, T_ICLASS)) {
 		    defined_class = RBASIC(defined_class)->klass;
 		}
 
-		if (!rb_obj_is_kind_of(cfp->self, rb_class_real(defined_class))) {
+		if (!rb_obj_is_kind_of(cfp->self, defined_class)) {
 		    val = vm_method_missing(th, id, recv, num, blockptr, NOEX_PROTECTED);
 		}
 		else {
@@ -697,10 +724,13 @@ vm_yield_with_cfunc(rb_thread_t *th, const rb_block_t *block,
 	blockarg = Qnil;
     }
 
-    vm_push_frame(th, 0, VM_FRAME_MAGIC_IFUNC,
+    vm_push_frame(th, (rb_iseq_t *)ifunc, VM_FRAME_MAGIC_IFUNC,
 		  self, (VALUE)block->dfp,
 		  0, th->cfp->sp, block->lfp, 1);
 
+    if (blockargptr) {
+	th->cfp->lfp[0] = GC_GUARDED_PTR((VALUE)blockargptr);
+    }
     val = (*ifunc->nd_cfnc) (arg, ifunc->nd_tval, argc, argv, blockarg);
 
     th->cfp++;
@@ -778,7 +808,7 @@ vm_yield_setup_block_args(rb_thread_t *th, const rb_iseq_t * iseq,
     int argc = orig_argc;
     const int m = iseq->argc;
     VALUE ary, arg0;
-    rb_num_t opt_pc = 0;
+    int opt_pc = 0;
 
     th->mark_stack_len = argc;
 
@@ -893,8 +923,9 @@ vm_invoke_block(rb_thread_t *th, rb_control_frame_t *reg_cfp, rb_num_t num, rb_n
     const rb_block_t *block = GET_BLOCK_PTR();
     rb_iseq_t *iseq;
     int argc = (int)num;
+    VALUE type = GET_ISEQ()->local_iseq->type;
 
-    if (GET_ISEQ()->local_iseq->type != ISEQ_TYPE_METHOD || block == 0) {
+    if ((type != ISEQ_TYPE_METHOD && type != ISEQ_TYPE_CLASS) || block == 0) {
 	rb_vm_localjump_error("no block given (yield)", Qnil, 0);
     }
     iseq = block->iseq;
@@ -1024,28 +1055,31 @@ vm_getspecial(rb_thread_t *th, VALUE *lfp, VALUE key, rb_num_t type)
 	    }
 	}
 	else {
-	    val = rb_reg_nth_match(type >> 1, backref);
+	    val = rb_reg_nth_match((int)(type >> 1), backref);
 	}
     }
     return val;
 }
 
 static NODE *
-vm_get_cref(const rb_iseq_t *iseq, const VALUE *lfp, const VALUE *dfp)
+vm_get_cref0(const rb_iseq_t *iseq, const VALUE *lfp, const VALUE *dfp)
 {
-    NODE *cref = 0;
-
     while (1) {
 	if (lfp == dfp) {
-	    cref = iseq->cref_stack;
-	    break;
+	    if (!RUBY_VM_NORMAL_ISEQ_P(iseq)) return NULL;
+	    return iseq->cref_stack;
 	}
 	else if (dfp[-1] != Qnil) {
-	    cref = (NODE *)dfp[-1];
-	    break;
+	    return (NODE *)dfp[-1];
 	}
 	dfp = GET_PREV_DFP(dfp);
     }
+}
+
+static NODE *
+vm_get_cref(const rb_iseq_t *iseq, const VALUE *lfp, const VALUE *dfp)
+{
+    NODE *cref = vm_get_cref0(iseq, lfp, dfp);
 
     if (cref == 0) {
 	rb_bug("vm_get_cref: unreachable");
@@ -1054,15 +1088,17 @@ vm_get_cref(const rb_iseq_t *iseq, const VALUE *lfp, const VALUE *dfp)
 }
 
 static NODE *
-vm_cref_push(rb_thread_t *th, VALUE klass, int noex)
+vm_cref_push(rb_thread_t *th, VALUE klass, int noex, rb_block_t *blockptr)
 {
     rb_control_frame_t *cfp = vm_get_ruby_level_caller_cfp(th, th->cfp);
-    NODE *cref = NEW_NODE_LONGLIFE(NODE_BLOCK, klass, 0, 0);
-    cref->nd_file = 0;
+    NODE *cref = NEW_BLOCK(klass);
     cref->nd_visi = noex;
 
-    if (cfp) {
-	cref->nd_next = (NODE *)rb_gc_write_barrier((VALUE)vm_get_cref(cfp->iseq, cfp->lfp, cfp->dfp));
+    if (blockptr) {
+	cref->nd_next = vm_get_cref0(blockptr->iseq, blockptr->lfp, blockptr->dfp);
+    }
+    else if (cfp) {
+	cref->nd_next = vm_get_cref0(cfp->iseq, cfp->lfp, cfp->dfp);
     }
 
     return cref;
@@ -1084,16 +1120,35 @@ vm_get_cbase(const rb_iseq_t *iseq, const VALUE *lfp, const VALUE *dfp)
     return klass;
 }
 
+static inline VALUE
+vm_get_const_base(const rb_iseq_t *iseq, const VALUE *lfp, const VALUE *dfp)
+{
+    NODE *cref = vm_get_cref(iseq, lfp, dfp);
+    VALUE klass = Qundef;
+
+    while (cref) {
+	if (!(cref->flags & NODE_FL_CREF_PUSHED_BY_EVAL) &&
+	    (klass = cref->nd_clss) != 0) {
+	    break;
+	}
+	cref = cref->nd_next;
+    }
+
+    return klass;
+}
+
 static inline void
 vm_check_if_namespace(VALUE klass)
 {
+    VALUE str;
     switch (TYPE(klass)) {
       case T_CLASS:
       case T_MODULE:
 	break;
       default:
+	str = rb_inspect(klass);
 	rb_raise(rb_eTypeError, "%s is not a class/module",
-		 RSTRING_PTR(rb_inspect(klass)));
+		 StringValuePtr(str));
     }
 }
 
@@ -1106,21 +1161,33 @@ vm_get_ev_const(rb_thread_t *th, const rb_iseq_t *iseq,
     if (orig_klass == Qnil) {
 	/* in current lexical scope */
 	const NODE *root_cref = vm_get_cref(iseq, th->cfp->lfp, th->cfp->dfp);
-	const NODE *cref = root_cref;
+	const NODE *cref;
 	VALUE klass = orig_klass;
 
+	while (root_cref && root_cref->flags & NODE_FL_CREF_PUSHED_BY_EVAL) {
+	    root_cref = root_cref->nd_next;
+	}
+	cref = root_cref;
 	while (cref && cref->nd_next) {
-	    klass = cref->nd_clss;
+	    if (cref->flags & NODE_FL_CREF_PUSHED_BY_EVAL) {
+		klass = Qnil;
+	    }
+	    else {
+		klass = cref->nd_clss;
+	    }
 	    cref = cref->nd_next;
 
 	    if (!NIL_P(klass)) {
 		VALUE am = 0;
+		st_data_t data;
 	      search_continue:
-		if (RCLASS_IV_TBL(klass) &&
-		    st_lookup(RCLASS_IV_TBL(klass), id, &val)) {
+		if (RCLASS_CONST_TBL(klass) &&
+		    st_lookup(RCLASS_CONST_TBL(klass), id, &data)) {
+		    val = ((rb_const_entry_t*)data)->value;
 		    if (val == Qundef) {
 			if (am == klass) break;
 			am = klass;
+			if (is_defined) return 1;
 			rb_autoload_load(klass, id);
 			goto search_continue;
 		    }
@@ -1137,8 +1204,10 @@ vm_get_ev_const(rb_thread_t *th, const rb_iseq_t *iseq,
 	}
 
 	/* search self */
-	klass = root_cref->nd_clss;
-	if (NIL_P(klass)) {
+	if (root_cref && !NIL_P(root_cref->nd_clss)) {
+	    klass = root_cref->nd_clss;
+	}
+	else {
 	    klass = CLASS_OF(th->cfp->self);
 	}
 
@@ -1152,10 +1221,10 @@ vm_get_ev_const(rb_thread_t *th, const rb_iseq_t *iseq,
     else {
 	vm_check_if_namespace(orig_klass);
 	if (is_defined) {
-	    return rb_const_defined_from(orig_klass, id);
+	    return rb_public_const_defined_from(orig_klass, id);
 	}
 	else {
-	    return rb_const_get_from(orig_klass, id);
+	    return rb_public_const_get_from(orig_klass, id);
 	}
     }
 }
@@ -1166,7 +1235,8 @@ vm_get_cvar_base(NODE *cref)
     VALUE klass;
 
     while (cref && cref->nd_next &&
-	   (NIL_P(cref->nd_clss) || FL_TEST(cref->nd_clss, FL_SINGLETON))) {
+	   (NIL_P(cref->nd_clss) || FL_TEST(cref->nd_clss, FL_SINGLETON) ||
+	    (cref->flags & NODE_FL_CREF_PUSHED_BY_EVAL))) {
 	cref = cref->nd_next;
 
 	if (!cref->nd_next) {
@@ -1182,16 +1252,22 @@ vm_get_cvar_base(NODE *cref)
     return klass;
 }
 
+
+#ifndef USE_IC_FOR_IVAR
+#define USE_IC_FOR_IVAR 1
+#endif
+
 static VALUE
 vm_getivar(VALUE obj, ID id, IC ic)
 {
-#if 1
+#if USE_IC_FOR_IVAR
     if (TYPE(obj) ==  T_OBJECT) {
 	VALUE val = Qundef;
 	VALUE klass = RBASIC(obj)->klass;
 
-	if (ic->ic_class == klass) {
-	    long index = ic->ic_index;
+	if (LIKELY(ic->ic_class == klass &&
+		   ic->ic_vmstat == GET_VM_STATE_VERSION())) {
+	    long index = ic->ic_value.index;
 	    long len = ROBJECT_NUMIV(obj);
 	    VALUE *ptr = ROBJECT_IVPTR(obj);
 
@@ -1207,11 +1283,12 @@ vm_getivar(VALUE obj, ID id, IC ic)
 
 	    if (iv_index_tbl) {
 		if (st_lookup(iv_index_tbl, id, &index)) {
-		    if (index < len) {
+		    if ((long)index < len) {
 			val = ptr[index];
 		    }
-		    ic->ic_class = RBASIC(obj)->klass;
-		    ic->ic_index = index;
+		    ic->ic_class = klass;
+		    ic->ic_value.index = index;
+		    ic->ic_vmstat = GET_VM_STATE_VERSION();
 		}
 	    }
 	}
@@ -1229,19 +1306,61 @@ vm_getivar(VALUE obj, ID id, IC ic)
 #endif
 }
 
+static void
+vm_setivar(VALUE obj, ID id, VALUE val, IC ic)
+{
+#if USE_IC_FOR_IVAR
+    if (!OBJ_UNTRUSTED(obj) && rb_safe_level() >= 4) {
+	rb_raise(rb_eSecurityError, "Insecure: can't modify instance variable");
+    }
+
+    rb_check_frozen(obj);
+
+    if (TYPE(obj) == T_OBJECT) {
+	VALUE klass = RBASIC(obj)->klass;
+	st_data_t index;
+
+	if (LIKELY(ic->ic_class == klass &&
+		   ic->ic_vmstat == GET_VM_STATE_VERSION())) {
+	    long index = ic->ic_value.index;
+	    long len = ROBJECT_NUMIV(obj);
+	    VALUE *ptr = ROBJECT_IVPTR(obj);
+
+	    if (index < len) {
+		ptr[index] = val;
+		return; /* inline cache hit */
+	    }
+	}
+	else {
+	    struct st_table *iv_index_tbl = ROBJECT_IV_INDEX_TBL(obj);
+
+	    if (iv_index_tbl && st_lookup(iv_index_tbl, (st_data_t)id, &index)) {
+		ic->ic_class = klass;
+		ic->ic_value.index = index;
+		ic->ic_vmstat = GET_VM_STATE_VERSION();
+	    }
+	    /* fall through */
+	}
+    }
+    rb_ivar_set(obj, id, val);
+#else
+    rb_ivar_set(obj, id, val);
+#endif
+}
+
 static inline const rb_method_entry_t *
 vm_method_search(VALUE id, VALUE klass, IC ic)
 {
     rb_method_entry_t *me;
 #if OPT_INLINE_METHOD_CACHE
-    if (LIKELY(klass == ic->ic_class) &&
-	LIKELY(GET_VM_STATE_VERSION() == ic->ic_vmstat)) {
-	me = ic->ic_method;
+    if (LIKELY(klass == ic->ic_class &&
+	GET_VM_STATE_VERSION() == ic->ic_vmstat)) {
+	me = ic->ic_value.method;
     }
     else {
 	me = rb_method_entry(klass, id);
 	ic->ic_class = klass;
-	ic->ic_method = me;
+	ic->ic_value.method = me;
 	ic->ic_vmstat = GET_VM_STATE_VERSION();
     }
 #else
@@ -1272,24 +1391,24 @@ vm_search_normal_superclass(VALUE klass, VALUE recv)
 }
 
 static void
-vm_search_superclass(rb_control_frame_t *reg_cfp, rb_iseq_t *ip,
+vm_search_superclass(rb_control_frame_t *reg_cfp, rb_iseq_t *iseq,
 		     VALUE recv, VALUE sigval,
 		     ID *idp, VALUE *klassp)
 {
     ID id;
     VALUE klass;
 
-    while (ip && !ip->klass) {
-	ip = ip->parent_iseq;
+    while (iseq && !iseq->klass) {
+	iseq = iseq->parent_iseq;
     }
 
-    if (ip == 0) {
+    if (iseq == 0) {
 	rb_raise(rb_eNoMethodError, "super called outside of method");
     }
 
-    id = ip->defined_method_id;
+    id = iseq->defined_method_id;
 
-    if (ip != ip->local_iseq) {
+    if (iseq != iseq->local_iseq) {
 	/* defined by Module#define_method() */
 	rb_control_frame_t *lcfp = GET_CFP();
 
@@ -1298,21 +1417,31 @@ vm_search_superclass(rb_control_frame_t *reg_cfp, rb_iseq_t *ip,
 	    rb_raise(rb_eRuntimeError, "implicit argument passing of super from method defined by define_method() is not supported. Specify all arguments explicitly.");
 	}
 
-	while (lcfp->iseq != ip) {
+	while (lcfp->iseq != iseq) {
+	    rb_thread_t *th = GET_THREAD();
 	    VALUE *tdfp = GET_PREV_DFP(lcfp->dfp);
 	    while (1) {
 		lcfp = RUBY_VM_PREVIOUS_CONTROL_FRAME(lcfp);
+		if (RUBY_VM_CONTROL_FRAME_STACK_OVERFLOW_P(th, lcfp)) {
+		    rb_raise(rb_eNoMethodError,
+			     "super called outside of method");
+		}
 		if (lcfp->dfp == tdfp) {
 		    break;
 		}
 	    }
 	}
 
-	id = lcfp->me->original_id;
+	/* temporary measure for [Bug #2420] [Bug #3136] */
+	if (!lcfp->me) {
+	    rb_raise(rb_eNoMethodError, "super called outside of method");
+	}
+
+	id = lcfp->me->def->original_id;
 	klass = vm_search_normal_superclass(lcfp->me->klass, recv);
     }
     else {
-	klass = vm_search_normal_superclass(ip->klass, recv);
+	klass = vm_search_normal_superclass(iseq->klass, recv);
     }
 
     *idp = id;
@@ -1341,6 +1470,11 @@ vm_throw(rb_thread_t *th, rb_control_frame_t *reg_cfp,
 
 	      search_parent:
 		if (cfp->iseq->type != ISEQ_TYPE_BLOCK) {
+		    if (cfp->iseq->type == ISEQ_TYPE_CLASS) {
+			cfp = RUBY_VM_PREVIOUS_CONTROL_FRAME(cfp);
+			dfp = cfp->dfp;
+			goto search_parent;
+		    }
 		    dfp = GC_GUARDED_PTR_REF((VALUE *) *dfp);
 		    base_iseq = base_iseq->parent_iseq;
 
@@ -1364,7 +1498,7 @@ vm_throw(rb_thread_t *th, rb_control_frame_t *reg_cfp,
 
 		    while ((VALUE *)cfp < th->stack + th->stack_size) {
 			if (cfp->dfp == dfp) {
-			    VALUE epc = epc = cfp->pc - cfp->iseq->iseq_encoded;
+			    VALUE epc = cfp->pc - cfp->iseq->iseq_encoded;
 			    rb_iseq_t *iseq = cfp->iseq;
 			    int i;
 
@@ -1406,21 +1540,36 @@ vm_throw(rb_thread_t *th, rb_control_frame_t *reg_cfp,
 	    else if (state == TAG_RETURN) {
 		rb_control_frame_t *cfp = GET_CFP();
 		VALUE *dfp = GET_DFP();
-		VALUE * const lfp = GET_LFP();
+		VALUE *lfp = GET_LFP();
+		int in_class_frame = 0;
 
 		/* check orphan and get dfp */
 		while ((VALUE *) cfp < th->stack + th->stack_size) {
+		    if (!lfp) {
+			lfp = cfp->lfp;
+		    }
+		    if (cfp->dfp == lfp && cfp->iseq->type == ISEQ_TYPE_CLASS) {
+			in_class_frame = 1;
+			lfp = 0;
+		    }
+
 		    if (cfp->lfp == lfp) {
 			if (VM_FRAME_TYPE(cfp) == VM_FRAME_MAGIC_LAMBDA) {
 			    VALUE *tdfp = dfp;
 
+			    if (in_class_frame) {
+				/* lambda {class A; ... return ...; end} */
+				dfp = cfp->dfp;
+				goto valid_return;
+			    }
+
 			    while (lfp != tdfp) {
 				if (cfp->dfp == tdfp) {
 				    /* in lambda */
 				    dfp = cfp->dfp;
 				    goto valid_return;
 				}
-				tdfp = GC_GUARDED_PTR_REF((VALUE *)*dfp);
+				tdfp = GC_GUARDED_PTR_REF((VALUE *)*tdfp);
 			    }
 			}
 		    }
@@ -1473,7 +1622,7 @@ vm_expandarray(rb_control_frame_t *cfp, VALUE ary, rb_num_t num, int flag)
     rb_num_t space_size = num + is_splat;
     VALUE *base = cfp->sp, *ptr;
     volatile VALUE tmp_ary;
-    long len;
+    rb_num_t len;
 
     if (TYPE(ary) != T_ARRAY) {
 	ary = rb_ary_to_ary(ary);
@@ -1483,11 +1632,11 @@ vm_expandarray(rb_control_frame_t *cfp, VALUE ary, rb_num_t num, int flag)
 
     tmp_ary = ary;
     ptr = RARRAY_PTR(ary);
-    len = RARRAY_LEN(ary);
+    len = (rb_num_t)RARRAY_LEN(ary);
 
     if (flag & 0x02) {
 	/* post: ..., nil ,ary[-1], ..., ary[0..-num] # top */
-	long i = 0, j;
+	rb_num_t i = 0, j;
 
 	if (len < num) {
 	    for (i=0; i<num-len; i++) {
@@ -1504,7 +1653,7 @@ vm_expandarray(rb_control_frame_t *cfp, VALUE ary, rb_num_t num, int flag)
     }
     else {
 	/* normal: ary[num..-1], ary[num-2], ary[num-3], ..., ary[0] # top */
-	int i;
+	rb_num_t i;
 	VALUE *bptr = &base[space_size - 1];
 
 	for (i=0; i<num; i++) {
@@ -1530,8 +1679,8 @@ vm_expandarray(rb_control_frame_t *cfp, VALUE ary, rb_num_t num, int flag)
 static inline int
 check_cfunc(const rb_method_entry_t *me, VALUE (*func)())
 {
-    if (me && me->type == VM_METHOD_TYPE_CFUNC &&
-	me->body.cfunc.func == func) {
+    if (me && me->def->type == VM_METHOD_TYPE_CFUNC &&
+	me->def->body.cfunc.func == func) {
 	return 1;
     }
     else {
@@ -1546,16 +1695,9 @@ inline
 VALUE
 opt_eq_func(VALUE recv, VALUE obj, IC ic)
 {
-    VALUE val = Qundef;
-
     if (FIXNUM_2_P(recv, obj) &&
 	BASIC_OP_UNREDEFINED_P(BOP_EQ)) {
-	if (recv == obj) {
-	    val = Qtrue;
-	}
-	else {
-	    val = Qfalse;
-	}
+	return (recv == obj) ? Qtrue : Qfalse;
     }
     else if (!SPECIAL_CONST_P(recv) && !SPECIAL_CONST_P(obj)) {
 	if (HEAP_CLASS_OF(recv) == rb_cFloat &&
@@ -1565,29 +1707,25 @@ opt_eq_func(VALUE recv, VALUE obj, IC ic)
 	    double b = RFLOAT_VALUE(obj);
 
 	    if (isnan(a) || isnan(b)) {
-		val = Qfalse;
-	    }
-	    else if (a == b) {
-		val = Qtrue;
-	    }
-	    else {
-		val = Qfalse;
+		return Qfalse;
 	    }
+	    return  (a == b) ? Qtrue : Qfalse;
 	}
 	else if (HEAP_CLASS_OF(recv) == rb_cString &&
 		 HEAP_CLASS_OF(obj) == rb_cString &&
 		 BASIC_OP_UNREDEFINED_P(BOP_EQ)) {
-	    val = rb_str_equal(recv, obj);
-	}
-	else {
-	    const rb_method_entry_t *me = vm_method_search(idEq, CLASS_OF(recv), ic);
-	    extern VALUE rb_obj_equal(VALUE obj1, VALUE obj2);
-
-	    if (check_cfunc(me, rb_obj_equal)) {
-		return recv == obj ? Qtrue : Qfalse;
-	    }
+	    return rb_str_equal(recv, obj);
 	}
     }
 
-    return val;
+    {
+	const rb_method_entry_t *me = vm_method_search(idEq, CLASS_OF(recv), ic);
+
+	if (check_cfunc(me, rb_obj_equal)) {
+	    return recv == obj ? Qtrue : Qfalse;
+	}
+    }
+
+    return Qundef;
 }
+
